import os
import sys
import json
import getopt

from jinja2 import Environment, PackageLoader, select_autoescape
from ryuso import util


HELP = '''RYU Sequential Orchestrator Factory

Command-line interface for building a working sequential orchestrator from a
RYU spec.

OPTIONS:
    -h --help
        Help. Print this message and exit.

    -o DIR
        Output directory. This is the directory in which RYUSO will create a
        subdirectory containing the package to be built. DIR defaults to ./

    -s SPEC_FILE
        The path to the RYU spec to be built. By default SPEC_FILE
        is ./ryu_spec.json

AUTHOR:
    Written by Marcus Belcastro

SEE ALSO:
    Source code: https://gitlab.com/delta1512/ryu-sequential-orchestrator
'''

spec_file = './ryu_spec.json'
out_dir = './'


opts, args = getopt.getopt(sys.argv[1:], 'ho:s:', ['help'])

for opt, arg in opts:
    if opt == '-h' or opt == '--help':
        print(HELP)
        sys.exit(0)
    elif opt == '-o':
        out_dir = arg
    elif opt == '-s':
        spec_file = arg


env = Environment(
    loader=PackageLoader('ryuso'),
    autoescape=select_autoescape()
)

# Fetch the RYU spec
with open(spec_file, 'r') as ryu_spec_f:
    ryu_spec_str = ryu_spec_f.read() # So that we can copy it later
    ryu_spec = json.loads(ryu_spec_str)

assert util.validate_spec(ryu_spec), 'The spec provided contains errors.'

pkg_name = ryu_spec.get('name').lower().replace(' ', '_')

# First build the directory structure
os.makedirs(os.path.join(out_dir, pkg_name, 'hooks'))
pkg_root = os.path.join(out_dir, pkg_name)

# Next build main
t = env.get_template('__main__.py.temp')

with open(os.path.join(pkg_root, '__main__.py'), 'w') as main_f:
    main_f.write(t.render(pkg_name=pkg_name))

# Then init
t = env.get_template('__init__.py.temp')
hook_files = set()

for h in ryu_spec.get('hooks'):
    # Get the second to last object in the reference (this is the .py file)
    # where the hooks will be contained.
    # TODO: Make this generalised so we can have any hierarchy
    ref = ryu_spec.get('hooks').get(h).get('reference').split('.')[-2]
    hook_files.add(ref)

with open(os.path.join(pkg_root, '__init__.py'), 'w') as main_f:
    main_f.write(t.render(pkg_name=pkg_name, hook_files=hook_files))

# Then build the hook stubs
t = env.get_template('func_defs.py.temp')

for hf in hook_files:
    with open(os.path.join(pkg_root, 'hooks', '{}.py'.format(hf)), 'w') as hfd:
        hfd.write(t.render(current_parent=hf, hooks=ryu_spec.get('hooks')))

# Add an init for the hooks
with open(os.path.join(pkg_root, 'hooks', '__init__.py'), 'w') as _:
    pass

# Finally, copy the spec over
with open(os.path.join(pkg_root, 'ryu_spec.json'), 'w') as ryu_spec_f:
    ryu_spec_f.write(ryu_spec_str)

print('Done.')



import time


def validate_argument(argument_name: str, arg_spec: dict):
    mandatory_keys = ('type', 'required')

    # Check that all the keys exist by doing a set subtraction
    assert set(arg_spec.keys()) & set(mandatory_keys) == set(mandatory_keys), 'Missing required keys in the hook {}'.format(argument_name)

    # Check the type of all the key's values
    assert isinstance(arg_spec.get('type'), str), 'Hook {} has malformed type, should be string'.format(argument_name)
    assert isinstance(arg_spec.get('required'), bool), 'Hook {} has malformed required, should be bool'.format(argument_name)
    assert isinstance(arg_spec.get('description', ''), str), 'Hook {} has malformed description, should be string'.format(argument_name)

    arg_type = arg_spec.get('type')
    assert arg_type in ('string', 'int', 'float', 'list', 'dict', 'bool'), 'Invalid type provided for argument {}'.format(argument_name)

    return True


def validate_hook(hook: str, hook_spec: dict, hooks: dict):
    mandatory_keys = ('reference', 'args')

    # Check that all the keys exist by doing a set subtraction
    assert set(hook_spec.keys()) & set(mandatory_keys) == set(mandatory_keys), 'Missing required keys in the hook {}'.format(hook)

    # Check the type of all the key's values
    assert isinstance(hook_spec.get('reference'), str), 'Hook {} has malformed reference, should be string'.format(hook)
    assert isinstance(hook_spec.get('args'), dict), 'Hook {} has malformed args, should be dict structure'.format(hook)
    assert isinstance(hook_spec.get('dependencies', []), list), 'Hook {} has malformed dependencies, should be list structure'.format(hook)
    assert isinstance(hook_spec.get('rectifiers', []), list), 'Hook {} has malformed rectifiers, should be list structure'.format(hook)
    assert isinstance(hook_spec.get('description', ''), str), 'Hook {} has malformed description, should be string'.format(hook)

    for arg in hook_spec.get('args'):
        validate_argument(arg, hook_spec.get('args').get(arg))

    all_hook_refs = hook_spec.get('dependencies', []) + hook_spec.get('rectifiers', [])
    for h in all_hook_refs:
        assert h.split('.')[-1] in hooks, 'Hook {} referenced a hook that does not exist (referencing {})'.format(hook, h)
        # A hook cannot reference itself
        assert h.split('.')[-1] != hook, 'Hook {} attempted to depend on itself'.format(hook)

    return True


def validate_program(program_name: str, program_spec: dict, hooks: tuple):
    mandatory_keys = ('hooks', 'args', 'outputs')

    # Check that all the keys exist by doing a set subtraction
    assert set(program_spec.keys()) & set(mandatory_keys) == set(mandatory_keys), 'Missing required keys in the program {}'.format(program_name)

    # Check the type of all the key's values
    assert isinstance(program_spec.get('hooks'), list), 'Program {} has malformed hooks, should be list structure'.format(program_name)
    assert isinstance(program_spec.get('args'), dict), 'Program {} has malformed args, should be dict structure'.format(program_name)
    assert isinstance(program_spec.get('outputs'), dict), 'Program {} has malformed outputs, should be dict structure'.format(program_name)
    assert isinstance(program_spec.get('description', ''), str), 'Program {} has malformed description, should be string'.format(program_name)

    for arg in program_spec.get('args'):
        validate_argument(arg, program_spec.get('args').get(arg))

    for h in program_spec.get('hooks'):
        assert h.split('.')[-1] in hooks, 'Program {} referenced a hook that does not exist (referencing {})'.format(program_name, h)

    return True


def validate_spec(ryu_spec: dict):
    mandatory_keys = ('name', 'programs', 'hooks')

    # Check that all the keys exist by doing a set subtraction
    assert set(ryu_spec.keys()) & set(mandatory_keys) == set(mandatory_keys), 'Missing required keys in the spec'

    hook_names = tuple(ryu_spec.get('hooks').keys())

    for p in ryu_spec.get('programs'):
        validate_program(p, ryu_spec.get('programs').get(p), hook_names)

    refs = set()
    for h in ryu_spec.get('hooks'):
        validate_hook(h, ryu_spec.get('hooks').get(h), hook_names)

        # Check for duplicate implementation references in hooks
        ref = ryu_spec.get('hooks').get(h).get('reference')
        assert not ref in refs, 'A duplicate hook implementation reference {} was found. Each hook must have a unique reference'.format(ref)
        refs.add(ref)

    return True


def compress_spec(ryu_spec: dict):
    # Remove non-mandatory output arguments
    to_remove_outputs = []

    for p in ryu_spec.get('programs'):
        for o in ryu_spec.get('programs').get(p).get('outputs'):
            if not ryu_spec.get('programs').get(p).get('outputs').get(o).get('required', False):
                to_remove_outputs.append(o)

        for o in to_remove_outputs:
            ryu_spec['programs'][p]['outputs'].pop(o)

    return ryu_spec


def get_std_task_metadata(task: dict):
    return {
        '_program_name' : task.get('program'),
        '_program_time_start' : int(time.time() * 1000)
    }


def has_failed(data_pool):
    return any([
        has_rectifier_failed(data_pool),
        has_dependency_failed(data_pool),
        has_program_failed(data_pool)
    ])


def has_rectifier_failed(data_pool):
    return data_pool.get('_rectifier_fail', False)


def has_dependency_failed(data_pool):
    return data_pool.get('_dependency_fail', False)


def has_program_failed(data_pool):
    return data_pool.get('_program_fail', False)



import json
import socket
import logging
import traceback

from ryuso import Orchestrator


MAX_BYTES_RECV = 100000000 # 100MB


class OrchestratorServer:
    '''
    OrchestratorServer: The server for a particular RYU spec

    This class takes a RYU Orchestrator and manages its external connections and
    requests.

    In its current state, this server is totally sequential and not concurrent.
    '''
    def __init__(self, orch: Orchestrator, host: str='0.0.0.0', port: int=60856, q: int=512):
        self.orchestrator = orch
        self.host = host
        self.port = port
        self.queue_len = q

    def do_run_task(self, task: dict):
        if self.orchestrator.validate_task(task):
            return self.orchestrator.run_task(task)

    def start_server(self):
        logging.info(
            'RYUSO server is starting with host %s, port %s, queue len %s',
            self.host,
            self.port,
            self.queue_len
        )
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s.bind((self.host, self.port))
        self.s.listen(self.queue_len)

        while True:
            has_sent = False
            conn, addr = self.s.accept()

            logging.info('Accepted connection from %s', addr)

            with conn:
                err = 'Unknown error' # Catch-all definition for finally statement
                try:
                    data, _adata, _flags, _addr = conn.recvmsg(MAX_BYTES_RECV)

                    task = json.loads(data.decode())

                    logging.debug('Server received task: %s', task)

                    result = self.do_run_task(task)

                    logging.debug('Task successful, results are: %s', result)

                    # Send the results back to the client
                    conn.sendto(json.dumps(result).encode(), addr)
                    has_sent = True

                    logging.debug('Successfully returned results to client')
                except UnicodeDecodeError:
                    err = 'Could not decode the message provided'
                    logging.error('Data could not be decoded: %s', data)
                except json.JSONDecodeError:
                    err = 'Message provided was not valid JSON'
                    logging.error('Bad JSON provided: %s', data)
                except KeyboardInterrupt:
                    has_sent = True
                    logging.warning('Caught SIGTERM on server during processing. The following task was processing: %s', task)
                    break
                except:
                    logging.critical('An unknown error occurred:\n%s', traceback.format_exc())
                    break
                finally:
                    if not has_sent:
                        conn.sendto(json.dumps({'_server_error' : err}).encode(), addr)
        logging.info('RYUSO successfulyl exited main loop')

    def __del__(self):
        self.s.close()
        logging.debug('Successfully closed the socket')



import pprint as pp
import logging
import traceback

from ryuso import util

class Hook:
    '''
    Hook: The data structure that manages a specific hook in the RYU spec.

    To fully build this object, you must init it and then add all of the
    rectification functions.
    '''
    def __init__(self, function, qualified_name: str):
        '''
        :param function: The callable function that is the hook.
        :param qualified_name: The qualified name that appears in the spec as
            hook.reference. This is so that the hook can be indexed in
            containing objects.
        '''
        self.function = function
        self.qualified_name = qualified_name
        self.base_name = qualified_name.split('.')[-1]
        self.is_rectifier = False
        self.is_dependency = False
        self.rectifiers = []
        self.dependencies = []

    def _catch_failure(func): #pylint: disable=E0213
        def catch_failure_inner_wrapper(self, data_pool: dict):
            try:
                return func(self, data_pool) #pylint: disable=E1102
            except KeyboardInterrupt:
                # Catch everything except for ctrl+c
                return data_pool
            except Exception as e:
                logging.warning(
                    'Hook %s failed to complete, traceback: %s',
                    self.qualified_name,
                    traceback.format_exc()
                )
                # Flag that runtime has failed (failed execution of program)
                data_pool['_program_fail'] = True
                # Flag that a rectifier failed. This means that all hope is lost
                data_pool['_rectifier_fail'] = self.is_rectifier
                data_pool['_dependency_fail'] = self.is_dependency
                data_pool['_last_failed_program'] = self.qualified_name
                return data_pool
        return catch_failure_inner_wrapper

    def _trace(func): #pylint: disable=E0213
        def trace_inner_wrapper(self, data_pool: dict):
            if not data_pool.get('_traceback', False):
                data_pool['_traceback'] = []
            data_pool['_traceback'].append(self.qualified_name)

            if '_step_through' in data_pool:
                print('IN DEBUGGING MODE, About to run {}'.format(self.qualified_name))
                print('IN DEBUGGING MODE, Dumping the data pool...')
                pp.pprint(data_pool)
                input('IN DEBUGGING MODE, press enter to step...')
            return func(self, data_pool) #pylint: disable=E1102
        return trace_inner_wrapper


    def add_rectifier(self, rectifier):
        rectifier.is_rectifier = True # Ensure that the hook to be added is a rectifier
        self.rectifiers.append(rectifier)

    def add_dependency(self, dependency):
        dependency.is_dependency = True # Ensure that the hook to be added is a dependency
        self.dependencies.append(dependency)

    def reverse_run(self, data_pool):
        logging.info('Rectification being run on %s', self.qualified_name)
        current_data_pool = data_pool
        for r in self.rectifiers:
            current_data_pool = r.run(current_data_pool)

            # If a rectifier failed, then fail everything
            if util.has_rectifier_failed(current_data_pool):
                logging.error('Rectification for %s failed at %s', self.qualified_name, r.qualified_name)
                return current_data_pool

        logging.debug('Rectification completed successfully')
        return current_data_pool

    def run_dependencies(self, data_pool):
        logging.debug('Running dependencies for %s', self.qualified_name)
        current_data_pool = data_pool

        for d in self.dependencies:
            current_data_pool = d.run(current_data_pool)

            if util.has_dependency_failed(current_data_pool):
                logging.error('Dependencies for %s failed at %s', self.qualified_name, d.qualified_name)
                return current_data_pool

        logging.debug('Dependencies completed successfully')
        return current_data_pool

    @_trace
    @_catch_failure
    def run(self, data_pool: dict):
        data_pool = self.run_dependencies(data_pool)

        if util.has_dependency_failed(data_pool):
            return data_pool

        return self.function(data_pool)

    def __del__(self):
        pass



import logging

from ryuso import Hook, util

class Orchestrator:
    '''
    Orchestrator: The result of a built RYU spec

    This class builds and implements all of the programs of the RYU spec.

    This class can directly take in a request and process it according to the spec.
    '''
    def __init__(self, ryu_spec: dict):
        assert util.validate_spec(ryu_spec)
        self.ryu_spec = util.compress_spec(ryu_spec)
        self.implementation = {}

        self.implementation['programs'] = {}
        self.implementation['hooks'] = {}

        for p in ryu_spec.get('programs'):
            # Each program consists of a list of keys for the hooks member
            # self.add_hook needs to be called to populate the hooks dictionary
            self.implementation['programs'][p] = [h.split('.')[-1] for h in ryu_spec.get('programs').get(p).get('hooks')]

    # STUB
    def validate_task(self, task: dict):
        '''
        Validates a task request to ensure that it is correct
        '''
        return True

    # STUB
    def validate_output(self, program_name: str, out_data: dict):
        '''
        Validates a task's output to see if it conforms to the spec
        '''
        return True

    def add_hook(self, hook: Hook):
        '''
        Adds a hook to the hook structure.
        '''
        self.implementation['hooks'][hook.base_name] = hook

    def load_runtime_for(self, program_name: str):
        '''
        Gets the list of function references for a particular program.
        The implementation structure stores programs and hook references and
        this simply dereferences that list of references.

        :returns: List of Hook objects representing the hooks of the
            program provided.
        '''
        runtime_hooks = self.implementation.get('programs').get(program_name)

        return [self.implementation.get('hooks').get(hook_ref) for hook_ref in runtime_hooks]

    def mask_output(self, program_name: str, out_data: dict):
        logging.debug('Raw program output: %s', out_data)
        # Get all the names of all the outputs for the program
        mask = set(self.ryu_spec.get('programs').get(program_name).get('outputs').keys())

        # Keep all debug info only if debugging is enabled or program unsuccessful
        if util.has_failed(out_data):
            debug_info_keys = filter(lambda k: k.startswith('_'), out_data)

            for k in debug_info_keys:
                mask.add(k)

        keys_to_remove = set(out_data.keys()) - mask

        for k in keys_to_remove:
            out_data.pop(k)

        return out_data

    def run_task(self, task: dict):
        '''
        Runs a single task. It will create a list of hooks to run and then run
        them based on the spec for the program provided.
        '''
        assert self.validate_task(task), 'The task requested is not valid'

        program = task.get('program', None)
        data_pool = task.get('data_pool', {})
        metadata = task.get('metadata', {})

        logging.info('Running program %s', program)

        # Metadata is added to the data pool as a magic data item
        if bool(metadata):
            logging.debug('Metadata: %s', metadata)

            for md in metadata:
                k = md if md.startswith('_') else '_' + md
                data_pool[k] = metadata.get(md)

        data_pool.update(util.get_std_task_metadata(task))

        runtime = self.load_runtime_for(program)

        logging.debug('Generated runtime: %s', runtime)

        for i, h in enumerate(runtime):
            data_pool = h.run(data_pool)

            # If a hook has flagged that it has failed, then begin the process
            # of rectification.
            if util.has_program_failed(data_pool):
                logging.info(
                    'Runtime fail detected in program %s, commencing rectification',
                    program
                )

                # From 1 before the erroneous hook, start executing rectification
                for h_idx in range(i - 1, -1, -1):
                    data_pool = runtime[h_idx].reverse_run(data_pool)

                    # If any rectifier fails, then stop doing everything
                    if util.has_rectifier_failed(data_pool):
                        logging.info(
                            'Program %s signalled failure at rectification for %s, stopping',
                            program,
                            runtime[h_idx].qualified_name
                        )
                        break
                break

        final = self.mask_output(program, data_pool)

        assert self.validate_output(program, final), 'The final output for the program is missing at least one item defined in the spec'

        return final

    def __del__(self):
        pass
